/* arch/arm/boot/compressed/tomtomgo-init.c
 *
 * Special initialization for TomTom GO.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <asm/arch/tomtomgo-init.h>

/* Hack several arch include files */
#include <asm/io.h>
#include <asm/arch/map.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/regs-serial.h>
#include <asm/arch/regs-dsc.h>

#undef  S3C2410_VA_GPIO
#undef  S3C2410_VA_CLKPWR
#undef  S3C2410_VA_UART0
#undef  S3C2410_VA_UART1
#undef  S3C2410_VA_UART2
#define S3C2410_VA_GPIO		S3C2410_PA_GPIO
#define S3C2410_VA_CLKPWR	S3C2410_PA_CLKPWR
#define S3C2410_VA_UART0	S3C2410_PA_UART0
#define S3C2410_VA_UART1	S3C2410_PA_UART1
#define S3C2410_VA_UART2	S3C2410_PA_UART2

#define S3C2410_UFCON0 		(S3C2410_PA_UART0 + S3C2410_UFCON)
#define S3C2410_UMCON0 		(S3C2410_PA_UART0 + S3C2410_UMCON)
#define S3C2410_ULCON0 		(S3C2410_PA_UART0 + S3C2410_ULCON)
#define S3C2410_UCON0  		(S3C2410_PA_UART0 + S3C2410_UCON)
#define S3C2410_UBRDIV0		(S3C2410_PA_UART0 + S3C2410_UBRDIV)

/* Chose baud rate and divisor calculation */
#define CURRENT_BAUD_RATE	115200
#define UBRDIV(pclk, baud)	(pclk / 16 / baud - 1)

/* Chosen CPU speeds for different models */
#define S3C2410_PLLSPEED 199200000
#define S3C2440_PLLSPEED 393000000 // 376800000 // 405600000

/* Dividers for S3C2410 */
#if S3C2410_PLLSPEED == 199200000
#define S3C2410_HDIVN	1
#define S3C2410_PDIVN	1
#define S3C2410_MDIV	0x9e
#define S3C2410_PDIV	3
#define S3C2410_SDIV	1
#define S3C2410_HCLKDIV	2
#define S3C2410_PCLKDIV	4
#elif S3C2410_PLLSPEED == 202000000
#define S3C2410_HDIVN	1
#define S3C2410_PDIVN	1
#define S3C2410_MDIV	0x5d
#define S3C2410_PDIV	1
#define S3C2410_SDIV	1
#define S3C2410_HCLKDIV	2
#define S3C2410_PCLKDIV	4
#elif S3C2410_PLLSPEED == 202800000
#define S3C2410_HDIVN	1
#define S3C2410_PDIVN	1
#define S3C2410_MDIV	0xa1
#define S3C2410_PDIV	3
#define S3C2410_SDIV	1
#define S3C2410_HCLKDIV	2
#define S3C2410_PCLKDIV	4
#elif S3C2440_PLLSPEED == 376800000
#define S3C2440_HDIVN	2
#define S3C2440_PDIVN	1
#define S3C2440_MDIV	0x95
#define S3C2440_PDIV	3
#define S3C2440_SDIV	1
#define S3C2440_HCLKDIV	4
#define S3C2440_PCLKDIV	8
#else
#error "Select a proper S3C2410 PLL speed"
#endif

#define S3C2410_PCLK	(S3C2410_PLLSPEED / S3C2410_PCLKDIV)

/* Dividers for S3C2440 */
#if S3C2440_PLLSPEED == 399651840
#define S3C2440_HDIVN	2
#define S3C2440_PDIVN	1
#define S3C2440_MDIV	0x6e
#define S3C2440_PDIV	3
#define S3C2440_SDIV	1
#define S3C2410_HCLKDIV	4
#define S3C2410_PCLKDIV	8
#elif S3C2440_PLLSPEED == 400000000
#define S3C2440_HDIVN	3
#define S3C2440_PDIVN	1
#define S3C2440_MDIV	0x5c
#define S3C2440_PDIV	1
#define S3C2440_SDIV	1
#define S3C2440_HCLKDIV	3
#define S3C2440_PCLKDIV	6
#elif S3C2440_PLLSPEED == 405600000
#define S3C2440_HDIVN	3
#define S3C2440_PDIVN	1
#define S3C2440_MDIV	0xa1
#define S3C2440_PDIV	3
#define S3C2440_SDIV	1
#define S3C2440_HCLKDIV	3
#define S3C2440_PCLKDIV	6
// 7bh, 2h,0h |393000000|393M|  3.0M| 0.1| 0.1| 0.3|0|
#elif S3C2440_PLLSPEED == 393000000
#define S3C2440_HDIVN	2
#define S3C2440_PDIVN	1
#define S3C2440_MDIV	0x7b
#define S3C2440_PDIV	2
#define S3C2440_SDIV	1
#define S3C2440_HCLKDIV	4
#define S3C2440_PCLKDIV	8
#else
#error "Select a proper S3C2440 PLL speed"
#endif

#define S3C2440_PCLK	(S3C2440_PLLSPEED / S3C2440_PCLKDIV)

void tomtomgo_init(void)
{
	unsigned long cpu;
	unsigned long clkdivn;
	unsigned long mpllcon;
	unsigned long ubrdiv;
	int i;

	/* Determine CPU type */
	cpu = __raw_readl(S3C2410_GSTATUS1);
	switch (cpu) {
	case 0x32410000:
	case 0x32410002:
		/* S3C2410 */
		clkdivn = (S3C2410_HDIVN<<1) | S3C2410_PDIVN;
		mpllcon = (S3C2410_MDIV<<12) | (S3C2410_PDIV<<4) | S3C2410_SDIV;
		ubrdiv = UBRDIV(S3C2410_PCLK, CURRENT_BAUD_RATE);
		break;
	case 0x32440000:
	case 0x32440001:
		/* S3C2440 */
		clkdivn = (S3C2440_HDIVN<<1) | S3C2440_PDIVN;
		mpllcon = (S3C2440_MDIV<<12) | (S3C2440_PDIV<<4) | S3C2440_SDIV;
		ubrdiv = UBRDIV(S3C2440_PCLK, CURRENT_BAUD_RATE);
		/* reduce the bus driver power to lower the noise floor */
		__raw_writel(0x7fffffff, S3C2440_DSC0);
		__raw_writel(0xffffffff, S3C2440_DSC1);
		break;
	default:
		/* Unknown CPU, just don't touch anything */
		return;
	}

	/* Setup all registers */
	__raw_writel(clkdivn, S3C2410_CLKDIVN); /* Clock divider */
	__raw_writel(mpllcon, S3C2410_MPLLCON); /* MPLL configuration */
	__raw_writel(0x0000,  S3C2410_UFCON0);  /* FIFO control register, FIFO disable */
	__raw_writel(0x0000,  S3C2410_UMCON0);  /* Modem control register, AFC disable */
	__raw_writel(0x0003,  S3C2410_ULCON0);  /* Line control register, 8N1 */
	__raw_writel(0x0245,  S3C2410_UCON0);   /* Control register */
	__raw_writel(ubrdiv,  S3C2410_UBRDIV0); /* Baud rate divisor register */

	/* Delay for a little while */
	for (i = 0; i < 100; ++i)
		;
}

/* EOF */
