/* arch/arm/mach-s3c2410/mach-tomtomgo.c
 *
 * Machine support for the TomTom GO, an all-in-one car navigation product.
 * See <http://www.tomtomgo.com/>.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/device.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach-types.h>

#include <asm/arch/regs-serial.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-irq.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/tomtomgo-irq.h>
#include <asm/arch/tomtomgo-map.h>
#include <barcelona/gotype.h>

#include "s3c2410.h"
#include "s3c2440.h"
#include "clock.h"
#include "devs.h"
#include "tomtomgo-devs.h"
#include "cpu.h"
#include "pm.h"

static struct map_desc tomtomgo_iodesc[] __initdata = {
	IODESC_ENT(SDI),
	IODESC_ENT(IIS),
	{ TOMTOMGO_VA_IDE, TOMTOMGO_PA_IDE, TOMTOMGO_SZ_IDE, MT_DEVICE },
};

#define UCON S3C2410_UCON_DEFAULT
#define ULCON S3C2410_LCON_CS8 | S3C2410_LCON_PNONE
#define UFCON S3C2410_UFCON_DEFAULT

static struct s3c2410_uartcfg tomtomgo_uartcfgs[] = {
	[0] = {
		.hwport	     = 0,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	},
	[1] = {
		.hwport	     = 1,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	},
	[2] = {
		.hwport	     = 2,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	}
};

/* Standard TomTom GO devices */

static struct platform_device *tomtomgo_devices[] __initdata = {
	&s3c_device_usb,
	&s3c_device_lcd,
	&s3c_device_wdt,
	&s3c_device_i2c,
	&s3c_device_iis,
	&s3c_device_rtc,
	&s3c_device_sdi,
	&tomtomgo_device_ide,
	&tomtomgo_device_hw,
	&tomtomgo_device_adc,
	&tomtomgo_device_acc,
	&tomtomgo_device_bat,
	&tomtomgo_device_gpio,
	&tomtomgo_device_gps,
	&tomtomgo_device_onoff,
	&tomtomgo_device_pwm,
	&tomtomgo_device_rc,
	&tomtomgo_device_sd,
	&tomtomgo_device_ts,
};

static struct s3c24xx_board tomtomgo_board __initdata = {
	.devices       = tomtomgo_devices,
	.devices_count = ARRAY_SIZE(tomtomgo_devices)
};

static void __init tomtomgo_map_io(void)
{
	s3c24xx_init_io(tomtomgo_iodesc, ARRAY_SIZE(tomtomgo_iodesc));
	s3c24xx_init_uarts(tomtomgo_uartcfgs, ARRAY_SIZE(tomtomgo_uartcfgs));
	s3c24xx_set_board(&tomtomgo_board);
	IO_Init();
}

static void __init tomtomgo_init_irq(void)
{
	s3c2410_init_irq();
}

#ifdef CONFIG_PM
/* cache functions from arch/arm/mm/proc-arm920.S */
extern void arm920_flush_kern_cache_all(void);

static void tomtomgo_power_off(void)
{
	unsigned long regs_save[16];

	/* set GSTATUS3 resume address to bootloader init */
	__raw_writel(0x8024, S3C2410_GSTATUS3);
	/* flush cache back to ram */
	arm920_flush_kern_cache_all();
	/* ack any outstanding external interrupts before we go to sleep */
	__raw_writel(__raw_readl(S3C2410_EINTPEND), S3C2410_EINTPEND);
	/* turn off clocks over sleep */
	__raw_writel(0x00, S3C2410_CLKCON);
	/* send the cpu to sleep... */
	s3c2410_cpu_suspend(regs_save);
	/* if this ever returns, just hang */
	for (;;)
		;
	/* NOTREACHED */
}

static void __init tomtomgo_init_machine(void)
{
	s3c2410_pm_init();
	pm_power_off = tomtomgo_power_off;
}
#endif /* CONFIG_PM */

MACHINE_START(TOMTOMGO, "TomTom GO")
     MAINTAINER("Dimitry Andric <dimitry.andric@tomtom.com>")
     BOOT_MEM(S3C2410_SDRAM_PA, S3C2410_PA_UART, S3C2410_VA_UART)
     BOOT_PARAMS(S3C2410_SDRAM_PA + 0x0)
     MAPIO(tomtomgo_map_io)
     INITIRQ(tomtomgo_init_irq)
     .timer		= &s3c2410_timer,
#ifdef CONFIG_PM
     INIT_MACHINE(tomtomgo_init_machine)
#endif /* CONFIG_PM */
MACHINE_END
