/* arch/arm/mach-s3c2410/tomtomgo-devs.h
 *
 * Definitions of TomTom GO platform devices.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/device.h>
#include <asm/hardware.h>
#include <asm/arch/tomtomgo-map.h>
#include <asm/arch/tomtomgo-irq.h>

/* TomTom GO IDE device */

static struct resource tomtomgo_ide_resource[] = {
	[0] = {
		.start = TOMTOMGO_PA_IDE,
		.end   = TOMTOMGO_PA_IDE + TOMTOMGO_SZ_IDE,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = TOMTOMGO_IRQ_IDE,
		.end   = TOMTOMGO_IRQ_IDE,
		.flags = IORESOURCE_IRQ,
	},
};

struct platform_device tomtomgo_device_ide = {
	.name          = "tomtomgo-ide",
	.id            = -1,
	.num_resources = ARRAY_SIZE(tomtomgo_ide_resource),
	.resource      = tomtomgo_ide_resource,
};

EXPORT_SYMBOL(tomtomgo_device_ide);


/* TomTom GO hardware detection device */

struct platform_device tomtomgo_device_hw = {
	.name          = "tomtomgo-hw",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_hw);


/* TomTom GO ADC device */

static struct resource tomtomgo_adc_resource[] = {
	[0] = {
		.start = S3C2410_PA_ADC,
		.end   = S3C2410_PA_ADC + S3C2410_SZ_ADC,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = IRQ_TC,
		.end   = IRQ_ADC,
		.flags = IORESOURCE_IRQ,
	},
};

struct platform_device tomtomgo_device_adc = {
	.name          = "tomtomgo-adc",
	.id            = -1,
	.num_resources = ARRAY_SIZE(tomtomgo_adc_resource),
	.resource      = tomtomgo_adc_resource,
};

EXPORT_SYMBOL(tomtomgo_device_adc);


/* TomTom GO accelerometer device */

struct platform_device tomtomgo_device_acc = {
	.name          = "tomtomgo-acc",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_acc);


/* TomTom GO battery device */

struct platform_device tomtomgo_device_bat = {
	.name          = "tomtomgo-bat",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_bat);


/* TomTom GO GPIO device */

struct platform_device tomtomgo_device_gpio = {
	.name          = "tomtomgo-gpio",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_gpio);


/* TomTom GO GPS device */

static struct resource tomtomgo_gps_resource[] = {
	[0] = {
		.start = TOMTOMGO_IRQ_GPS_PPS,
		.end   = TOMTOMGO_IRQ_GPS_PPS,
		.flags = IORESOURCE_IRQ,
	},
};

struct platform_device tomtomgo_device_gps = {
	.name          = "tomtomgo-gps",
	.id            = -1,
	.num_resources = ARRAY_SIZE(tomtomgo_gps_resource),
	.resource      = tomtomgo_gps_resource,
};

EXPORT_SYMBOL(tomtomgo_device_gps);


/* TomTom GO on/off-button device */

static struct resource tomtomgo_onoff_resource[] = {
	[0] = {
		.start = TOMTOMGO_IRQ_ONOFF,
		.end   = TOMTOMGO_IRQ_ONOFF,
		.flags = IORESOURCE_IRQ,
	},
};

struct platform_device tomtomgo_device_onoff = {
	.name          = "tomtomgo-onoff",
	.id            = -1,
	.num_resources = ARRAY_SIZE(tomtomgo_onoff_resource),
	.resource      = tomtomgo_onoff_resource,
};

EXPORT_SYMBOL(tomtomgo_device_onoff);


/* TomTom GO PWM device */

struct platform_device tomtomgo_device_pwm = {
	.name          = "tomtomgo-pwm",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_pwm);


/* TomTom GO remote control device */

struct platform_device tomtomgo_device_rc = {
	.name          = "tomtomgo-rc",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_rc);


/* TomTom GO storage device */

struct platform_device tomtomgo_device_sd = {
	.name          = "tomtomgo-sd",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_sd);


/* TomTom GO touchscreen */

struct platform_device tomtomgo_device_ts = {
	.name          = "tomtomgo-ts",
	.id            = -1,
};

EXPORT_SYMBOL(tomtomgo_device_ts);

/* EOF */
