/* linux/arch/arm/mach-s3c2410/wake.c
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Thomas Kleffel <tk@maintech.de>
 *
 * S3C2410 Power Manager Advanced Wakeup Management
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/kobject.h>
#include <linux/sysfs.h>

#define WAKE_SLOTS (10)

extern struct subsystem power_subsys;

struct wake_module_ops {
	void* (*create_context)(void);
	void  (*free_context)(void *);

	ssize_t (*get_param)(void *, struct attribute *, char *);
	ssize_t (*set_param)(void *, struct attribute *, const char *, size_t);

#define	WAKEUP_NONE	0
#define WAKEUP_SLEEP	1
#define	WAKEUP_WAKE	2
#define WAKEUP_RESET	3

	void  (*before_sleep)(void *);
	int   (*decide_wakeup)(void *);
	void  (*after_wakeup)(void *);
};

struct wake_module {
	struct kobject		kobj;

	char			*name;
	struct wake_module_ops 	 ops;
	struct attribute_group	*attributes;
};

struct wake_slot {
	struct kobject	 	 kobj;
	struct kobject		 mparam;

	struct wake_module	*module;
	void			*mdata;
	int 		 	 active;
};

struct wake_t {
	struct subsystem	subsys;
	struct kset	 	slotset;
	struct kset	 	moduleset;
	struct kset		mparamset;
	struct wake_slot	slots[WAKE_SLOTS];
};

extern int s3c2410_wake_module_register(struct wake_module *module);
extern void s3c2410_wake_module_unregister(struct wake_module *module);
extern int s3c2410_wake_init(void);
extern void s3c2410_wake_before_sleep(void);
extern void s3c2410_wake_after_wakeup(void);
extern int s3c2410_wake_decide_wakeup(void);

