/*
*  linux/drivers/ide/arm/tomtom.c - TomTomGo harddisk temperature test
*
*  Copyright (C) 2004 TomTom BV
*  Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*/

#include <linux/delay.h>
#include <barcelona/gotype.h>

static unsigned long ideBaseAddress;

// IDE interface
#define rIDEDATAREAD     (*(volatile unsigned short*)(ideBaseAddress+(0*4)))
#define rIDEDATAWRITE    (*(volatile unsigned short*)(ideBaseAddress+(0*4)))
#define rIDEERROR        (*(volatile unsigned char*) (ideBaseAddress+(1*4)))
#define rIDEFEATURES     (*(volatile unsigned char*) (ideBaseAddress+(1*4)))
#define rIDESECTORCOUNT  (*(volatile unsigned char*) (ideBaseAddress+(2*4)))
#define rIDESECTORNUMBER (*(volatile unsigned char*) (ideBaseAddress+(3*4)))
#define rIDECYLINDERLOW  (*(volatile unsigned char*) (ideBaseAddress+(4*4)))
#define rIDECYLINDERHIGH (*(volatile unsigned char*) (ideBaseAddress+(5*4)))
#define rIDEDEVICEHEAD   (*(volatile unsigned char*) (ideBaseAddress+(6*4)))
#define rIDESTATUS       (*(volatile unsigned char*) (ideBaseAddress+(7*4)))
#define rIDECOMMAND      (*(volatile unsigned char*) (ideBaseAddress+(7*4)))

/* bits in Status register: */
#define BSY 0x0080

static void IDE_WaitForBusyCleared(void)
{
	while ((rIDESTATUS & BSY) > 0);
}

static void IDE_DoCommand(unsigned short cCommand)
{
	IDE_WaitForBusyCleared();
	rIDECOMMAND = cCommand;
	IDE_WaitForBusyCleared();
}

static void IDE_InitDisk(void)
{
	IO_Activate(HDD_PWR_ON);
	IO_SetInput(HDD_IRQ);
	IO_SetInput(HD_LED);
	IO_Activate(AT_RST);
	mdelay(100);
	IO_Deactivate(AT_RST);
	IO_Activate(HDD_BUF_EN);
	mdelay(100);
	IDE_WaitForBusyCleared();
}

static void IDE_Off(void)
{
	IO_Deactivate(HDD_BUF_EN);
	IO_Activate(AT_RST);
	IO_Deactivate(HDD_PWR_ON);
	IO_Activate(HD_LED);
	IO_Deactivate(HDD_IRQ);
}

static int IDE_GetTemperature(void)
{
	rIDEDEVICEHEAD = 0xA0;
	rIDEFEATURES = 0xDF;
	IDE_DoCommand(0xEF);
	return (rIDESECTORNUMBER);
}

void tomtom_hddtemp_setaddress(unsigned long address)
{
	ideBaseAddress = address;
}

int tomtom_hddtemp_check(void)
{
	int temperature;

	IDE_InitDisk();
	temperature = IDE_GetTemperature();
	IDE_Off();
	if ((temperature < 2) || (temperature > 58)) return -1; else return 0;
}
