#!/bin/sh -e


# Default settings
verfile="tomtom.ver"
defcfg="config-default"
rebuild="n"
makeflags=""
makedefs="V=0"
makejobs="-j 4"

# Environment for building
export ARCH="arm"
export CROSS_COMPILE="arm-linux-"

usage() {
  echo "Usage: $0 {rebuild|silent|verbose|single}"
  exit 1
}

get_version() {
  source ${verfile}
  if test -z "${TTVER}"; then
    echo "$0: ERROR: Invalid version in ${verfile}!"
    exit 1
  fi

  read -p "Update kernel build number? [n] "
  if test "${REPLY}" = "y"; then
    echo "**** Checking if you edited any sources..."
    if ! cvs -Q diff > /dev/null 2>&1; then
      echo "$0: ERROR: You need to commit your changes first."
      exit 1
    fi
    echo "**** Checking if your sources are up to date..."
    if ! cvs -Qn commit > /dev/null 2>&1; then
      echo "$0: ERROR: You need to update your sources first."
      exit 1
    fi
    TTVER=`expr ${TTVER} + 1`
    sed -i "s/^TTVER=.*$/TTVER=${TTVER}/" ${verfile}
    echo "**** Committing new build number ${TTVER}..."
    cvs -Q commit -m "New kernel build ${TTVER}."
    echo "**** Tagging with new build number ${TTVER}..."
    cvs -Q tag build-tt${TTVER}
  fi
}

make_clean() {
  echo "**** Cleaning...."
  nice make ${makeflags} ${makedefs} distclean
}

make_config() {
  echo "**** Setting configuration..."
  cp ${defcfg} .config
  nice make ${makeflags} ${makedefs} silentoldconfig
}

make_build() {
  echo "**** Building kernel..."
  nice make ${makeflags} ${makejobs} ${makedefs}
}

# Main starts here
while test -n "$1"; do
  case "$1" in
  rebuild)
    rebuild="y"
    ;;
  silent)
    makeflags="-s"
    makedefs="V=0"
    ;;
  verbose)
    makeflags=""
    makedefs="V=1"
    ;;
  single)
    makejobs=""
    ;;
  *)
    usage
    ;;
  esac
  shift
done

get_version

# Set up our local version suffix
echo "-tt${TTVER}" > localversion

echo "**** Building kernel #${TTVER}..."

if test "${rebuild}" = "y"; then
  make_clean
  make_config
elif test ! -f .config; then
  make_config
fi

make_build

echo "**** Successfully built kernel #${TTVER}:"
ls -l arch/arm/boot/zImage

# EOF
