/* CoolSound engine. Copyright (c) TomTom 2004-2005.
 *
 * 
 * Copyright (c) 2005 TomTom BV
 * Written by: Thomas Kleffel <tk@maintech.de> 
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/completion.h> 
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>

#include <asm/uaccess.h>
#include <asm/irq.h>

#include <barcelona/Barc_snd.h>
#include <barcelona/gotype.h>

#include "driver.h"
#include "iis.h"
#include "dma.h"


#include "wm8711.h"
#include "wm8971.h"

#if 0
#define SANITY_CHECK \
	do { \
		if (destination == NULL || source == NULL || samples == 0 || samples > 10000) { \
			printk(KERN_WARNING "%s: Unsane input, dest=%p, src=%p, smp=%zu\n", __FUNCTION__, destination, source, samples); \
			return; \
		} \
	} while (0)
#else
#define SANITY_CHECK do {} while (0)
#endif

#define SOUND_MAJOR 244
#define CSB_MAX 16

static char sound_name[] = "tomtomgo-sound";
static unsigned int do_filter;

static struct fasync_struct *fasync_queue;

static struct iis_attached_ops iis_attached_ops;

static struct cs_buffer csb[CSB_MAX];

static void magic_tasklet(unsigned long data);
DECLARE_TASKLET(magic_tasklet_handle, magic_tasklet, 0L);


struct cs_primary_buffer_description priInputDesc = {
	.channels = 2,
	.bits_per_sample = 16,
	.frequency = 22050,
};

struct cs_primary_buffer_description priOutputDesc = {
	.channels = 2,
	.bits_per_sample = 16,
	.frequency = 22050,
};

static int setup_primixer(struct cs_buffer *, struct cs_primary_buffer_description *);

static void pmx_copy_1(void *destination, void *source, size_t samples, unsigned int flags) {
	SANITY_CHECK;
	memcpy(destination, source, samples);
}

static void pmx_copy_2(void *destination, void *source, size_t samples, unsigned int flags) {
	SANITY_CHECK;
	memcpy(destination, source, samples<<1);
}

static void pmx_copy_4(void *destination, void *source, size_t samples, unsigned int flags) {
	SANITY_CHECK;
	memcpy(destination, source, samples<<2);
}

#if 0

static void pmx_zero_1(void *destination, size_t samples) {
	memset(destination, 0, samples<<0);
}

static void pmx_zero_2(void *destination, size_t samples) {
	memset(destination, 0, samples<<1);
}

static void pmx_zero_4(void *destination, size_t samples) {
	memset(destination, 0, samples<<2);
}

#endif

static void pmx_mono_16_playmix(void *destination, void *source, size_t samples, unsigned int flags) {
	int16_t *d, *s;
	int32_t res;
	
	SANITY_CHECK;

	//If we are the first buffer to be mixed in - just copy;
	if(flags & PMXF_MIXIN_FIRST) {
		pmx_copy_2(destination, source, samples, flags);
		return;
	}
	
	
	d=(int16_t *) destination;
	s=(int16_t *) source;
	
	while(samples--) {
		res = d[0] + s[0];
		if(res < -0x7FFF) res = -0x7FFF;
		else if(res > 0x7FFF) res = 0x7FFF;
		d[0] = (int16_t) res;
		
		d++; s++;
	}
}

static void pmx_stereo_16_playmix(void *destination, void *source, size_t samples, unsigned int flags) {
	//Like mono - only twice the samples...
	SANITY_CHECK;
	pmx_mono_16_playmix(destination, source, samples << 1, flags);
}

static void pmx_mono_to_stereo_16_playback(void *destination, void *source, size_t samples, unsigned int flags) {
	int16_t *d, *s;
	
	SANITY_CHECK;

	s = (int16_t *) source;
	d = (int16_t *) destination;
	
	while(samples--) {
		*(d++) = *(s);
		*(d++) = *(s++);
	}
}

static void pmx_mono_to_stereo_16_playmix(void *destination, void *source, size_t samples, unsigned int flags) {
	int16_t *d, *s;
	int32_t res;

	SANITY_CHECK;

	s = (int16_t *) source;
	d = (int16_t *) destination;
	
	if(flags & PMXF_MIXIN_FIRST) {
		return pmx_mono_to_stereo_16_playback(destination, source, samples, flags);
	}
		
	while(samples--) {
		res = d[0] + s[0];
		if(res < -0x7FFF) res = -0x7FFF;
		else if(res > 0x7FFF) res = 0x7FFF;
		d[0] = (int16_t) res;
		
		res = d[1] + s[0];
		if(res < -0x7FFF) res = -0x7FFF;
		else if(res > 0x7FFF) res = 0x7FFF;
		d[1] = (int16_t) res;
		
		d+=2; s++;
	}
}

static void pmx_mono_to_stereo_16_record(void *destination, void *source, size_t samples, unsigned int flags) {
	int16_t *d, *s;

	SANITY_CHECK;
	
	s = (int16_t *) source;
	d = (int16_t *) destination;

	while(samples--) {
		//*(d++) = (s[0]>>1) + (s[1]>>1);
		// Maybe this is mono to mono, with one silent channel?
		// Thomas, I know, dirty hack, sorry :-) This basically
		// maps mono disquised as stereo to mono.
		*(d++) = (s[0]) + (s[1]);
		s+=2;
	}
}

static int csb_selected;
void cs_do_magic(void);

int csb_init(void) {
	int i;
	
	for(i=0;i<CSB_MAX;i++) {
		csb[i].no = i;
		csb[i].state = CS_BUFFER_FREE;
	}
	
	return 0;
}

int csb_acquire(void) {
	unsigned long flags;
	int i, res;
	
	res = -EBUSY;
	
	local_irq_save(flags);
		for(i=0;i<CSB_MAX;i++) {
			printk("cs: state %d %d\n",csb[i].state, CS_BUFFER_FREE);
			if(csb[i].state == CS_BUFFER_FREE) {
				csb[i].state = CS_BUFFER_INITIALIZING;
				res = i; break;
			}
		}
	local_irq_restore(flags);
	
	return res;
}

int csb_run_in_out(struct cb_in_out *cbio) {
	unsigned long flags;
	
	int res = 0;
        int csb_in_no;
        int csb_out_no;

	if(!cbio) return -EINVAL;

	csb_in_no = cbio->in;
	csb_out_no = cbio->out;

	if((csb_out_no < 0) || (csb_out_no >= CSB_MAX)) {
		printk("CoolSound: Warning: trying to run invalid buffer handle (%d)\n", csb_out_no);
		return -EINVAL;
	}		

	local_irq_save(flags);
		if(csb[csb_in_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: trying to run free buffer (%d)\n", csb_in_no);
			res = -EBUSY;
			goto csb_run_out;
		}
	
		if(csb[csb_in_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: trying to run initializing buffer  (%d)\n", csb_in_no);
			res = -EBUSY;
			goto csb_run_out;
		}

		if(csb[csb_out_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: trying to run free buffer (%d)\n", csb_out_no);
			res = -EBUSY;
			goto csb_run_out;
		}
	
		if(csb[csb_out_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: trying to run initializing buffer  (%d)\n", csb_out_no);
			res = -EBUSY;
			goto csb_run_out;
		}
		
		
		if(csb[csb_out_no].desc.type == CSB_TYPE_PLAYBACK) {		
			if((res = setup_primixer(&csb[csb_out_no], &priOutputDesc))) {
				printk("CoolSound: Error: Could not find a suitable playback primixer for buffer %i.\n", csb_out_no);
				res = -EBUSY;
				goto csb_run_out;
			}
			
			if(csb[csb_out_no].samples < (priOutputDesc.chunksize * priOutputDesc.chunks)) {
				printk("CoolSound: Error: Can't run buffer %i with less (%u) samples than primary buffer(%u).\n",
					csb_out_no, csb[csb_out_no].samples, (priOutputDesc.chunksize * priOutputDesc.chunks));
				res = -EBUSY;
				goto csb_run_out;
			}

			if(!sound_dma_output_ready()) {
				printk("CoolSound: Error: Can't run secondary playback buffer %d when primary buffer is not ready.\n",
					csb_out_no);
				res = -EBUSY;
				goto csb_run_out; 
			}
		}
				
		if(csb[csb_in_no].desc.type == CSB_TYPE_RECORD) {
			if((res = setup_primixer(&csb[csb_in_no], &priInputDesc))) {
				printk("CoolSound: Error: Could not find a suitable record primixer for buffer %i.\n", csb_in_no);
				goto csb_run_out;
			}	
		
			if(csb[csb_in_no].samples < (priInputDesc.chunksize * priInputDesc.chunks)) {
				printk("CoolSound: Error: Can't run buffer %i with less (%u) samples than primary buffer(%u).\n",
					csb_in_no, csb[csb_in_no].samples, (priInputDesc.chunksize * priInputDesc.chunks));
				res = -EBUSY;
				goto csb_run_out;
			}				
		
			if(!sound_dma_input_ready()) {
				printk("CoolSound: Error: Can't run secondary record buffer %d when primary buffer is not ready.\n",
					csb_in_no);
				res = -EBUSY;
				goto csb_run_out; 
			}
		}

		int in = 0;
		int out = 0;
		csb[csb_in_no].state = CS_BUFFER_RUNNING;
		csb[csb_out_no].state = CS_BUFFER_RUNNING;
		if(!sound_dma_input_running()) {
			printk("CoolSound: running input buffer.\n");
			sound_dma_input_reset();
                        in = 1;
                }

		if(!sound_dma_output_running()) {
			printk("CoolSound: running output buffer.\n");
			sound_dma_output_reset();
			out = 1;
		}

		if ((in) || (out)) {
			cs_do_magic();	
		}

		if (in) {
			iis_control_cmd(IIS_ALIGN_RX);
			
			if((res = sound_dma_input_start())) {
				printk("CoolSound: Error: could not start primary input buffer (%d).\n", res);
				goto csb_run_out;
				//TODO: bring back state machine to sane state.
			}
			
			iis_control_cmd(IIS_ACTIVE_RX);
		}

		if (out) {
			iis_control_cmd(IIS_ALIGN_TX);
							
			if((res = sound_dma_output_start())) {
				printk("CoolSound: Error: could not start primary input buffer (%d).\n", res);
				goto csb_run_out;
				//TODO: bring back state machine to sane state.
			}

			iis_control_cmd(IIS_ACTIVE_TX);
		}

csb_run_out:
	local_irq_restore(flags);
	
	return res;
}


int csb_run(int csb_no) {
	unsigned long flags;
	
	int res = 0;
	if((csb_no < 0) || (csb_no >= CSB_MAX)) {
		printk("CoolSound: Warning: trying to run invalid buffer handle (%d)\n", csb_no);
		return -EINVAL;
	}		
	
	local_irq_save(flags);
		if(csb[csb_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: trying to run free buffer (%d)\n", csb_no);
			res = -EBUSY;
			goto csb_run_out;
		}
	
		if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: trying to run initializing buffer  (%d)\n", csb_no);
			res = -EBUSY;
			goto csb_run_out;
		}
		
		
		if(csb[csb_no].desc.type == CSB_TYPE_PLAYBACK) {		
			if((res = setup_primixer(&csb[csb_no], &priOutputDesc))) {
				printk("CoolSound: Error: Could not find a suitable playback primixer for buffer %i.\n", csb_no);
				res = -EBUSY;
				goto csb_run_out;
			}
			
			if(csb[csb_no].samples < (priOutputDesc.chunksize * priOutputDesc.chunks)) {
				printk("CoolSound: Error: Can't run buffer %i with less (%u) samples than primary buffer(%u).\n",
					csb_no, csb[csb_no].samples, (priOutputDesc.chunksize * priOutputDesc.chunks));
				res = -EBUSY;
				goto csb_run_out;
			}

			if(!sound_dma_output_ready()) {
				printk("CoolSound: Error: Can't run secondary playback buffer %d when primary buffer is not ready.\n",
					csb_no);
				res = -EBUSY;
				goto csb_run_out; 
			}
				
			csb[csb_no].state = CS_BUFFER_RUNNING;
			if(!sound_dma_output_running()) {
				printk("CoolSound: running output buffer.\n");
				sound_dma_output_reset();
				cs_do_magic();	
				iis_control_cmd(IIS_ALIGN_TX);
								
				if((res = sound_dma_output_start())) {
					printk("CoolSound: Error: could not start primary input buffer (%d).\n", res);
					goto csb_run_out;
					//TODO: bring back state machine to sane state.
				}

				
				iis_control_cmd(IIS_ACTIVE_TX);
			} else {
				printk("CoolSound: output buffer already running.\n");
			}
		} else if(csb[csb_no].desc.type == CSB_TYPE_RECORD) {
			if((res = setup_primixer(&csb[csb_no], &priInputDesc))) {
				printk("CoolSound: Error: Could not find a suitable record primixer for buffer %i.\n", csb_no);
				goto csb_run_out;
			}	
		
			if(csb[csb_no].samples < (priInputDesc.chunksize * priInputDesc.chunks)) {
				printk("CoolSound: Error: Can't run buffer %i with less (%u) samples than primary buffer(%u).\n",
					csb_no, csb[csb_no].samples, (priInputDesc.chunksize * priInputDesc.chunks));
				res = -EBUSY;
				goto csb_run_out;
			}				
		
			if(!sound_dma_input_ready()) {
				printk("CoolSound: Error: Can't run secondary record buffer %d when primary buffer is not ready.\n",
					csb_no);
				res = -EBUSY;
				goto csb_run_out; 
			}
			
			csb[csb_no].state = CS_BUFFER_RUNNING;
			if(!sound_dma_input_running()) {
				printk("CoolSound: running input buffer.\n");
				sound_dma_input_reset();
				cs_do_magic();	
				iis_control_cmd(IIS_ALIGN_RX);
				
				if((res = sound_dma_input_start())) {
					printk("CoolSound: Error: could not start primary input buffer (%d).\n", res);
					goto csb_run_out;
					//TODO: bring back state machine to sane state.
				}
				
				iis_control_cmd(IIS_ACTIVE_RX);
			} else {
				printk("CoolSound: input buffer already running.\n");
			}
		} else {
			printk("CoolSound: Warning: trying to buffer (%d) of unknown type %d.\n", csb_no, csb[csb_no].desc.type);
		}
		

csb_run_out:
	local_irq_restore(flags);
	
	return res;
}

int csb_stop(int csb_no) {
	unsigned long flags;
	int res = 0;
	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;
	
	local_irq_save(flags);
		if(csb[csb_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: tried to stop free buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_stop_out;
		}
	
		if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: tried to stop initializing buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_stop_out;
		}
		
		csb[csb_no].state = CS_BUFFER_STOPPED;
		printk("CoolSound: Buffer %d now stopped.\n",csb_no);
		
csb_stop_out:
	local_irq_restore(flags);
	return res;
}

void csb_free(int csb_no) {
	if((csb_no < 0) || (csb_no >= CSB_MAX)) return;
	
	csb_stop(csb_no);
	
        printk("set buffer %d to CS_FREE_BUFFER\n",csb_no);
	csb[csb_no].state = CS_BUFFER_FREE;
}

int csb_queue_stop(struct cb_qstop *cbq) {
	unsigned long flags;
	int res = 0;
	int csb_no;
	
	if(!cbq) return -EINVAL;
	csb_no = cbq->buffer;

	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;
	
	local_irq_save(flags);
		if(csb[csb_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: tried to queue stop on free buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_queue_stop_out;
		}
	
		if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: tried to queue stop on initializing buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_queue_stop_out;
		}
		
		if((cbq->stop_position < 0) || (cbq->stop_position >= csb[csb_no].size)) {
			printk("CoolSound: Warning: tried to queue stop on buffer [%d] on out-of-limits position [%d]. Size of buffer is [%d].\n",
				csb_no, cbq->stop_position, csb[csb_no].size);
			res = -EINVAL;
			goto csb_queue_stop_out;
		}
		
		csb[csb_no].state = CS_BUFFER_STOP_PENDING;
		csb[csb_no].spos  = cbq->stop_position;
		printk("CoolSound: Buffer %d now stops at position %d. Current position is %d.\n", csb_no, csb[csb_no].spos, csb[csb_no].pointer);

csb_queue_stop_out:
	local_irq_restore(flags);
	return res;
}

int csb_setpos(int csb_no, int pos) {
	unsigned long flags;
	int res = 0;
	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;
	
	local_irq_save(flags);
		if(csb[csb_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: tried to setpos free buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_setpos_out;
		}
	
		if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: tried to setpos initializing buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_setpos_out;
		}
		
		if(csb[csb_no].size <= pos) { 
			printk("CoolSound: Warning: tried to setpos %d with pos(%d) bigger than size (%d).\n",csb_no, pos, csb[csb_no].size);
			res = -EBUSY;
			goto csb_setpos_out;
		}
		
		csb[csb_no].pointer = pos;

csb_setpos_out:
	local_irq_restore(flags);
	return res;

}

int csb_getpos(int csb_no) {
	int res = 0;
	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;

	if(csb[csb_no].state == CS_BUFFER_FREE) {
		printk("CoolSound: Warning: tried to getpos free buffer %d.\n",csb_no);
		res = -EBUSY;
		goto csb_getpos_out;
	}

	if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
		printk("CoolSound: Warning: tried to getpos initializing buffer %d.\n",csb_no);
		res = -EBUSY;
		goto csb_getpos_out;
	}
	
	res = csb[csb_no].pointer;
	
csb_getpos_out:
	return res;
}

int csb_setcbfreq(int csb_no, int cbfrequency) {
	unsigned long flags;
	int res = 0;
	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;

	local_irq_save(flags);
		if(csb[csb_no].state == CS_BUFFER_FREE) {
			printk("CoolSound: Warning: tried to setcbfreq free buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_setcbfreq_out;
		}
	
		if(csb[csb_no].state == CS_BUFFER_INITIALIZING) {
			printk("CoolSound: Warning: tried to setcbfreq initializing buffer %d.\n",csb_no);
			res = -EBUSY;
			goto csb_setcbfreq_out;
		}
	
		csb[csb_no].cbfreq = cbfrequency;
		
csb_setcbfreq_out:
	local_irq_restore(flags);
	return res;
}
	
static int sound_free_buffer(int csb_no) {
	struct cs_buffer *my_csb;

	if((csb_no < 0) || (csb_no >= CSB_MAX)) return -EINVAL;
	my_csb = &csb[csb_no];
	if (my_csb->buffer)
		dma_free_coherent(NULL, my_csb->size, my_csb->buffer, (dma_addr_t ) my_csb->phys);
	csb_free(csb_no);
	return 0;
}

static int sound_request_buffer(struct cs_buffer_description *bds) {
	int my_csb_no;
	struct cs_buffer *my_csb;
	
	if(!bds) return -EINVAL;
	
	//Fixed for now...
	//Don't forget to change size calculation!!!
	if(bds->frequency == 0) 	return -EINVAL;
	if(bds->channels<1)		return -EINVAL;
	if(bds->channels>2)		return -EINVAL;
	if(bds->bits_per_sample != 16)	return -EINVAL;
	
	if((bds->samples == 0) || (bds->samples > (1024 * 1024))) return -EINVAL;
	
	my_csb_no = csb_acquire();
	if(my_csb_no < 0) return my_csb_no;
	
	my_csb = &csb[my_csb_no];
		
	
	//TODO: dma_alloc_coherent seems to be bogus because we don't do DMA on those buffers
	my_csb->desc 	= (*bds);
	my_csb->bytes_per_sample = 2 * bds->channels;
	my_csb->size	= bds->samples * my_csb->bytes_per_sample;
	my_csb->samples	= bds->samples;
	my_csb->buffer	= (void *) dma_alloc_coherent(NULL, my_csb->size, (dma_addr_t *) &my_csb->phys, GFP_USER);
	my_csb->pointer	= 0;
		
	if(!my_csb->buffer) {
		printk("CoolSound: failed to allocate dma buffer\n");
		csb_free(my_csb_no);
		return -EINVAL;
	}
	
	my_csb->state	= CS_BUFFER_STOPPED;
	
	csb_selected = my_csb_no;
	return my_csb_no;
}

void cs_signal(int buffer) {
	static int z=1234;
	//printk("cs_signal(%d)\n", z);
	kill_fasync(&fasync_queue, SIGIO, z);
	z++;
}

void cs_do_magic(void) {
	int i, smp;
	int inbuffers, outbuffers;
	int do_more_magic;

	static int zeros = 0;
			
do_more_magic_here:

	inbuffers = 0;
	outbuffers = 0;
	do_more_magic = 0;
	
	
	
	//Care for output buffer
	smp = sound_dma_output_free();
	for(i=0;i<CSB_MAX;i++) {
		int is_rest = 0;
		int optr, nptr;
		
		if(csb[i].state == CS_BUFFER_FREE) {
			continue;
		}
		
		if(csb[i].state == CS_BUFFER_INITIALIZING) {
			continue;
		}	
		if(csb[i].desc.type != CSB_TYPE_PLAYBACK) {
			continue;
		}	
		
		if((csb[i].state == CS_BUFFER_RUNNING) || (csb[i].state == CS_BUFFER_STOP_PENDING)) {
			//printk("Mixing in %d.\n",i);
		} else continue; 
		
		if(csb[i].state == CS_BUFFER_STOP_PENDING) {
			int remaining;
			if(csb[i].pointer <= csb[i].spos) 	remaining = csb[i].spos - csb[i].pointer;
			else					remaining = csb[i].size - (csb[i].pointer - csb[i].spos);
			
			//Align to full samples;
			remaining /= csb[i].bytes_per_sample;
			remaining *= csb[i].bytes_per_sample;
			
			if(remaining <= (smp * csb[i].bytes_per_sample)) {
				smp = remaining / csb[i].bytes_per_sample;
				is_rest = 1;
			}
		}
	
		zeros = 0;
		optr = csb[i].pointer;
/*		
		printk("do_magic(output): csb:%u pointer:%ub size:%ub smp:%us fillsize:%ub, fillspace:%u\n",
			i, csb[i].pointer, csb[i].size, smp, (smp * csb[i].bytes_per_sample), (csb[i].size - csb[i].pointer));
*/					
		if((csb[i].size - csb[i].pointer) >= (smp * csb[i].bytes_per_sample)) {
			//printk("sound_dma_output_put(%p, %u, %u)\n", csb[i].buffer + csb[i].pointer, smp, 0);
			sound_dma_output_put(csb[i].buffer + csb[i].pointer, smp, 0, csb[i].primixer, outbuffers?0:PMXF_MIXIN_FIRST);
			csb[i].pointer+=smp * csb[i].bytes_per_sample;
		} else {
			int csmp = (csb[i].size - csb[i].pointer) / csb[i].bytes_per_sample;
			//printk("sound_dma_output_put(%p, %u, %u)\n", csb[i].buffer + csb[i].pointer, csmp, 0);
			sound_dma_output_put(csb[i].buffer + csb[i].pointer, csmp, 0, csb[i].primixer, outbuffers?0:PMXF_MIXIN_FIRST);
			//printk("sound_dma_output_put(%p, %u, %u)\n", csb[i].buffer, (smp-csmp), csmp);
			sound_dma_output_put(csb[i].buffer, (smp-csmp), csmp, csb[i].primixer, outbuffers?0:PMXF_MIXIN_FIRST);
			csb[i].pointer = (smp-csmp) * csb[i].bytes_per_sample;
		}

		nptr = csb[i].pointer;
			
		//Check if pointer went over notification...
		if(csb[i].cbfreq) {
			int notify_usermode = 0;
			
			if(nptr < optr) notify_usermode = 1;
			else if((smp * csb[i].bytes_per_sample) >= csb[i].cbfreq) notify_usermode = 1;
			else if((nptr%csb[i].cbfreq) < (optr%csb[i].cbfreq)) notify_usermode = 1;
			
			if(notify_usermode) cs_signal(i);
		}
		
		if(csb[i].state == CS_BUFFER_STOP_PENDING) {
			if(is_rest) {
				//TODO: Perhaps notify usermode of stopped buffer
				csb[i].state = CS_BUFFER_STOPPED;		
				do_more_magic = 1;
				printk("CoolSound: do_magic(): secondary buffer %d executed queued stop at position %d.\n", i, csb[i].spos);
			}
		}	
			
		outbuffers++;
	}

#if 1
	//printk("outbuffers=%d\n", outbuffers);
	if(!outbuffers) {
		int running, size;

		zeros += smp;
		//printk("CoolSound: do_magic(): sound_dma_output_zeroes(%u, 0)\n", smp);
		sound_dma_output_zeroes(smp, 0);
		
		running = sound_dma_output_running();
		size = sound_dma_output_size() >> 2;
		//printk("DEBUG: running=%d, zeros=%d, size=%d\n", running, zeros, size);
		if(running && zeros >= size) {
			printk("Stop Primary Playback!\n");
			iis_control_cmd(IIS_IDLE_TX);
			sound_dma_output_stop();
			iis_control_cmd(IIS_FLUSH_TX);
		}		
	}
#else
	/* TEMPORARY HACK: Forget zeroing out anything, until the suspend
	 * function has a way to wait for us to finish. */
	if(!outbuffers && sound_dma_output_running()) {
		printk("Stop Primary Playback!\n");
		iis_control_cmd(IIS_IDLE_TX);
		sound_dma_output_stop();
		iis_control_cmd(IIS_FLUSH_TX);
	}
#endif

	sound_dma_output_advance(smp);	
	
	//Do black filter magic
	if(do_filter) sound_dma_filter_magic();
	
	//care for input driver
	smp = sound_dma_input_filled();
	
	for(i=0;i<CSB_MAX;i++) {
		if(csb[i].state == CS_BUFFER_FREE) {
			continue;
		}
		
		if(csb[i].state == CS_BUFFER_INITIALIZING) {
			continue;
		}	
 		
		if(csb[i].desc.type != CSB_TYPE_RECORD) {
			continue;
		}	
		
		if(csb[i].state == CS_BUFFER_RUNNING) {
			int nptr;
			int optr = csb[i].pointer;
						
			if((csb[i].size - csb[i].pointer) >= smp * csb[i].bytes_per_sample) {
				sound_dma_input_get(csb[i].buffer + csb[i].pointer, smp, 0, csb[i].primixer, 0);
				csb[i].pointer+=smp * csb[i].bytes_per_sample;
			} else { 
				int csmp = ((csb[i].size - csb[i].pointer) / csb[i].bytes_per_sample);
				sound_dma_input_get(csb[i].buffer + csb[i].pointer, csmp, 0, csb[i].primixer, 0);	
				sound_dma_input_get(csb[i].buffer, (smp-csmp), csmp, csb[i].primixer, 0);
				csb[i].pointer = (smp-csmp) * csb[i].bytes_per_sample;
			}
	
			nptr = csb[i].pointer;
				
			if(csb[i].cbfreq) {
				int notify_usermode = 0;
				
				if(nptr < optr) notify_usermode = 1;
				else if((smp*csb[i].bytes_per_sample) >= csb[i].cbfreq) notify_usermode = 1;
				else if((nptr%csb[i].cbfreq) < (optr%csb[i].cbfreq)) notify_usermode = 1;
						
				if(notify_usermode) cs_signal(i);
			}
			inbuffers++;
		} 
		break; 
	}
	sound_dma_input_advance(smp);
	if(sound_dma_input_running() && (!inbuffers)) {
		printk("Stop Primary Record!\n");
		iis_control_cmd(IIS_IDLE_RX);
		sound_dma_input_stop();
		iis_control_cmd(IIS_FLUSH_RX);
	}
	
	if(do_more_magic) goto do_more_magic_here;
}

int sound_setup_primary(struct cs_primary_buffer_description *cpbd) {
	unsigned long flags;
	int res = 0;
	
	if(!cpbd) return -EINVAL;
	if(cpbd->bits_per_sample != 16) return -EINVAL;
	if(cpbd->channels != 2) return -EINVAL;
	
	if((cpbd->chunksize < 16 ) || (cpbd->chunksize > 4096 )) return -EINVAL;
	if((cpbd->chunks < 3) || (cpbd->chunks > 16)) return -EINVAL;
	
	
	local_irq_save(flags);
		if(sound_dma_output_running() || sound_dma_input_running()) {
                        printk("sound_setup_primary: dma still busy\n");
			res = -EBUSY;
			goto sound_setup_primary_out;
		}	
		
		if((res = sound_dma_setup_buffers(cpbd->chunksize * 4, cpbd->chunks))) {
                        printk("sound_setup_primary: setup buffers\n");
			goto sound_setup_primary_out;
		}
		
		if((res = iis_set_samplerate(cpbd->frequency))) {
                        printk("sound_setup_primary: set_samplerate\n");
			goto sound_setup_primary_out;
		}
		
		priInputDesc.frequency = cpbd->frequency;
		priOutputDesc.frequency = cpbd->frequency;
	
sound_setup_primary_out:	
	local_irq_restore(flags);
	return res;
}

static int setup_primixer(struct cs_buffer *pSecondary, struct cs_primary_buffer_description *pPrimaryDescription) {
	struct cs_buffer_description *pSD;
	struct cs_primary_buffer_description *pPD;
	
	if(!pSecondary) 		return -EINVAL;
	if(!pPrimaryDescription)	return -EINVAL;
	
	pSD = &pSecondary->desc;
	pPD = pPrimaryDescription; 
	
	if(!(pSD && pPD)) return -EINVAL;
	
	if((pSD->bits_per_sample != 8) && (pSD->bits_per_sample != 16)) {
		printk("CoolSound: Warning: setup_primixer(): Can't deal with %d bits per sample in secondary buffer.\n", pSD->bits_per_sample);
		return -EBUSY;
	}

	if((pSD->channels != 1) && (pSD->channels != 2)) {
		printk("CoolSound: Warning: setup_primixer(): Can't deal with %d channels in secondary buffer.\n", pSD->channels);
		return -EBUSY;
	}
		
	//We can't do 8<->16 Bit conversion by now
	if(pSD->bits_per_sample != pPD->bits_per_sample) {
		printk("CoolSound: Warning: setup_primixer(): Can't convert from secondary %d bits/sample to primary %d bits/sample.\n", pSD->bits_per_sample, pPD->bits_per_sample);
		return -EBUSY;
	}
	
	//We can't do sampling frequency conversion by now
	if(pSD->frequency != pPD->frequency) {
		printk("CoolSound: Warning: setup_primixer(): Can't convert from secondary %d smp/s to primary %d smp/s.\n", pSD->frequency, pPD->frequency);
		return -EBUSY;
	}
	
	if(pSD->channels == pPD->channels) {
		int bps = (pSD->bits_per_sample / 8) * pSD->channels;
		printk("CoolSound: Info: setup_primixer(): Primary buffer format equals secondary buffer format(%d bit, %d channels, %d Hz).\n",
			pSD->bits_per_sample, pSD->channels, pSD->frequency);
		
		if((pSD->bits_per_sample == 16) && (pSD->channels == 1) && (pSD->type == CSB_TYPE_PLAYBACK)) {
			printk("CoolSound: Info: setup_primixer(): Choose 'pmx_mono_16_playmix()' as primixer function.\n");
			pSecondary->primixer = pmx_mono_16_playmix;
			return 0;
		}

		if((pSD->bits_per_sample == 16) && (pSD->channels == 2) && (pSD->type == CSB_TYPE_PLAYBACK)) {
			printk("CoolSound: Info: setup_primixer(): Choose 'pmx_stereo_16_playmix()' as primixer function.\n");
			pSecondary->primixer = pmx_stereo_16_playmix;
			return 0;
		}
				
		switch(bps) {
			case 1:
				printk("CoolSound: Info: setup_primixer(): Choose 'pmx_copy_1()' as primixer function.\n");
				pSecondary->primixer = pmx_copy_1;
				return 0;
			
			case 2:
				printk("CoolSound: Info: setup_primixer(): Choose 'pmx_copy_2()' as primixer function.\n");
				pSecondary->primixer = pmx_copy_2;
				return 0;

			case 4:
				printk("CoolSound: Info: setup_primixer(): Choose 'pmx_copy_4()' as primixer function.\n");
				pSecondary->primixer = pmx_copy_4;
				return 0;
				
			default:
				printk("CoolSound: Warning: setup_primixer(): Don't have copy function for strange %d bytes per sample.\n", bps);
				return -EBUSY;
		}
						 
	} else if((pSD->channels == 1) && (pPD->channels == 2)) {
		if(pSD->type == CSB_TYPE_PLAYBACK) {
			printk("CoolSound: Info: setup_primixer(): Choose 'pmx_mono_to_stereo_16_playback()' as primixer function.\n");
			pSecondary->primixer = pmx_mono_to_stereo_16_playmix;
			return 0;
		} else if(pSD->type == CSB_TYPE_RECORD) {
			printk("CoolSound: Info: setup_primixer(): Choose 'pmx_mono_to_stereo_16_record()' as primixer function.\n");
			pSecondary->primixer = pmx_mono_to_stereo_16_record;
			return 0;
		} else {
			printk("CoolSound: Warning: setup_primixer(): Unknown buffer type %d.\n", (int) pSD->type);
			return -EBUSY;	
		}
	} else {		
		printk("CoolSound: Warning: setup_primixer(): Can't convert from secondary %d channels to primary %d channels.\n", pSD->channels, pPD->channels);
		return -EBUSY;	
	}
}


static void magic_tasklet(unsigned long data) {
	unsigned long flags;
	
	local_irq_save(flags);
	cs_do_magic();
	local_irq_restore(flags);
}


void cs_callback(enum eDMAPurpose purpose, int chunk) {
	tasklet_schedule(&magic_tasklet_handle);
}


/* open: Not much here, just check that there isn't anyone
 * using the device.
 */
static int sound_open(struct inode *inode, struct file *fp)
{
	trace();

	if (!(fp->f_mode & FMODE_WRITE || fp->f_mode & FMODE_READ))
		return -ENODEV;

	return 0;
}

static int sound_fasync(int fd, struct file *fp, int mode)
{
	return fasync_helper(fd, fp, mode, &fasync_queue);
}

static int sound_release(struct inode *inode, struct file *fp)
{
	trace();

	if (!(fp->f_mode & FMODE_WRITE || fp->f_mode & FMODE_READ))
		return -ENODEV;

	sound_fasync(-1, fp, 0);

	return 0;
}

void usermode_dbgcall(unsigned long arg) {
	printk("BARCSND: Usermode debug [%lu].\n", arg);
	switch(arg) {
		case 1: iis_showbus("usermode debug"); return;
		case 2: sound_dma_usermode_debug(); return;
		default:
			D(printk("barc_snd: unknown usermode debug request %lu.\n",arg);)
			break;
	}
}

static int sound_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg) {
	unsigned long flags;


	if (_IOC_TYPE(cmd) != COOLSOUND_DRIVER_MAGIC) return -ENOTTY;

	if (_IOC_DIR(cmd) & _IOC_WRITE)
		if (!access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd))) return -EFAULT;

	switch(cmd) {
	
	case COOLSOUND_SETUP_PRIMARY: {
		struct cs_primary_buffer_description cpbd;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cs_primary_buffer_description))) return -EFAULT;
		if(copy_from_user(&cpbd, (void *)arg, sizeof(struct cs_primary_buffer_description))) return -EFAULT;
		return sound_setup_primary(&cpbd);	
	}
	
	case COOLSOUND_GET_BUFFER: {
		struct cs_buffer_description cbd;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cs_buffer_description))) return -EFAULT;
		if(copy_from_user(&cbd, (void *)arg, sizeof(struct cs_buffer_description))) return -EFAULT;
		return sound_request_buffer(&cbd);
	}
	
	case COOLSOUND_PUT_BUFFER: {
		return sound_free_buffer(arg);
	}
	
	case COOLSOUND_SEL_BUFFER: {
		if((arg<0) || (arg>=CSB_MAX)) return -EINVAL;
		csb_selected = arg;
		break;
	}
		 
	case COOLSOUND_RUN_IN_OUT: {
		struct cb_in_out cbio;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cb_in_out))) return -EFAULT;
		if(copy_from_user(&cbio, (void *)arg, sizeof(struct cb_in_out))) return -EFAULT;
		return csb_run_in_out(&cbio);
	}

	case COOLSOUND_RUN_BUFFER: {
		return csb_run(arg);
	}
	
	case COOLSOUND_STOP_BUFFER: {
		return csb_stop(arg);
	}
	
	case COOLSOUND_QUEUE_STOP: {
		struct cb_qstop cqs;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cb_qstop))) return -EFAULT;
		if(copy_from_user(&cqs, (void *)arg, sizeof(struct cb_qstop))) return -EFAULT;
		return csb_queue_stop(&cqs);
	}
			
	case COOLSOUND_GETPOS: {
		return csb_getpos(arg);
	}

	case COOLSOUND_SETPOS: {
		struct cb_chpos chp;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cb_chpos))) return -EFAULT;
		if(copy_from_user(&chp, (void *)arg, sizeof(struct cb_chpos))) return -EFAULT;
		return csb_setpos(chp.buffer, chp.position);
	}
	
	case COOLSOUND_SETCBFREQ: {
		struct cb_cbfreq cbf;
		if(!access_ok(VERIFY_READ, (void *)arg, sizeof(struct cb_cbfreq))) return -EFAULT;
		if(copy_from_user(&cbf, (void *)arg, sizeof(struct cb_cbfreq))) return -EFAULT;
		return csb_setcbfreq(cbf.buffer, cbf.frequency);
	}

	case COOLSOUND_SETFILTER: {
		local_irq_save(flags);
		if(arg) {
			sound_dma_filter_init();
			
			do_filter=1;
			sound_dma_filter_resync();
		} else {
			do_filter=0;
			sound_dma_filter_shutdown();
		}
		local_irq_restore(flags);
		break;
	}

	case COOLSOUND_MUTE_INTERNAL:
		IO_Deactivate(AMP_ON);
		return 0;
		
	case COOLSOUND_UNMUTE_INTERNAL:
		IO_Activate(AMP_ON);
		return 0;

	case COOLSOUND_MUTE_EXTERNAL:
		IO_Activate(MUTE_EXT);
		return 0;
		
	case COOLSOUND_UNMUTE_EXTERNAL:
		IO_Deactivate(MUTE_EXT);
		return 0;

	case COOLSOUND_SET_VOLUME:
		if(iis_attached_ops.set_volume)
			return iis_attached_ops.set_volume(arg);
		else	return -EINVAL;
	
	case COOLSOUND_SET_VOLUME_INTERNAL:
		if(iis_attached_ops.set_volume_internal)
			return iis_attached_ops.set_volume_internal(arg);
		else	return -EINVAL;
	
	case COOLSOUND_SET_VOLUME_EXTERNAL:
		if(iis_attached_ops.set_volume_external)
			return iis_attached_ops.set_volume_external(arg);
		else	return -EINVAL;
	
	case COOLSOUND_SET_TREBLE:
		if(iis_attached_ops.set_treble)
			return iis_attached_ops.set_treble(arg);
		else	return -EINVAL;
	
	case COOLSOUND_SET_BASS:
		if(iis_attached_ops.set_bass)
			return iis_attached_ops.set_bass(arg);
		else	return -EINVAL;
	
	case COOLSOUND_SET_REG:
		if(iis_attached_ops.set_register)
			return iis_attached_ops.set_register(arg);
		else	return -EINVAL;

	/* for debugging purposes only */
	case COOLSOUND_DEBUG:
		usermode_dbgcall(arg);
		return 0;
	}
	return -ENOTTY;
}

static int sound_mmap(struct file *fp, struct vm_area_struct *vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long size = vma->vm_end - vma->vm_start;
 	unsigned long buf_phys = 0;
	size_t buf_size = 0;
	struct cs_buffer *my_csb = &csb[csb_selected];

	trace();
	/* Check if offset is zero, we don't allow mmapping at an offset. */
	if (offset) return -ENXIO;
	/* Don't allow copy on write mappings, for writing they must be shared. */
	if ((vma->vm_flags & VM_WRITE) && !(vma->vm_flags & VM_SHARED)) return -EINVAL;
	/* Don't allow a read and write mapping. */
	
	vma->vm_flags |= VM_LOCKED;
		
	
	
	if((my_csb->state == CS_BUFFER_FREE) || (my_csb->state == CS_BUFFER_INITIALIZING)) {
		printk("CoolSound: Warning: trying to mmap free or uninitialized buffer.\n");
		return -EBUSY;
	}
	
	buf_phys = (unsigned long) my_csb->phys;
	buf_size = my_csb->size;
	

	if(!(buf_size && buf_phys)) return -EINVAL;

	printk("CoolSound: MMapping [%i] Size:%u Phys:%p\n",csb_selected, buf_size, (void *)buf_phys);
	
	if(size > buf_size) {
		printk("CoolSound: Warning: trying to mmap area[%lu] bigger than primary buffer[%u].\n", size, buf_size);
		//return -ENXIO;
	}
	
	if(size < buf_size) {
		printk("CoolSound: Warning: mmapping area[%lu] smaller than primary buffer[%u].\n", size, buf_size);
	}

	if (remap_pfn_range(vma, vma->vm_start,
	                    buf_phys >> PAGE_SHIFT,
	                    size,
	                    vma->vm_page_prot))
		return -ENXIO;
	return 0;
}

static struct file_operations sound_fops = {
	.owner = THIS_MODULE,
	.open = sound_open,
	.release = sound_release,
	.ioctl = sound_ioctl,
	.mmap = sound_mmap,
	.fasync = sound_fasync
};

static void sound_poweron(void)
{
	IO_Deactivate(AMP_ON);
	IO_Deactivate(MUTE_INT);
	IO_Activate(DAC_PWR_ON);
	msleep(50);

	IO_SetFunction(I2SSDO);
	IO_SetFunction(CDCLK);
	IO_SetFunction(I2SSCLK);
	IO_SetFunction(I2SLRCK);
	IO_SetFunction(I2SSDI);

	IO_Activate(MUTE_INT);
	msleep(50);

	iis_init();

	switch (IO_GetCpuType()) {
		case GOCPU_S3C2410:
			wm8711_init(&iis_attached_ops);
			break;
		case GOCPU_S3C2440:
			wm8971_init(&iis_attached_ops);
			break;
	}
}

static void sound_poweroff(void)
{
	IO_Deactivate(AMP_ON);
	IO_Deactivate(MUTE_INT);
	IO_Deactivate(DAC_PWR_ON);
}

static int sound_probe(struct device *dev)
{
	int ret = 0;

	if((ret = sound_dma_init(dev, cs_callback))) {
		printk("Unable to associate primary DMA buffers. Error:%d\n", ret);
		return ret;
	}
	
	if((ret = sound_dma_setup_buffers(1024, 4))) {
		printk("Unable to allocate primary DMA buffers. Error:%d\n", ret);
		return ret;
	}
	
	D(printk("Registering chardev %s (%p)\n", sound_name, dev));
	ret = register_chrdev(SOUND_MAJOR, sound_name, &sound_fops);
	if (ret < 0) {
		printk("Unable to register chardev on major=%d (%d)\n", SOUND_MAJOR, ret);
		return ret;
	}

	sound_poweron();

	return 0;
}


static int sound_remove(struct device *dev)
{
	tasklet_kill(&magic_tasklet_handle);
	
	D(printk("Unregistering chardev\n"));
	sound_poweroff();
	
	return unregister_chrdev(SOUND_MAJOR, sound_name);
}

static void sound_shutdown(struct device *dev)
{
	D(printk("Shutting down\n"));
	sound_poweroff();
	/* Nothing yet */
}

#ifdef CONFIG_PM

static int sound_suspend(struct device *dev, u32 state, u32 level)
{
	D(printk("dev = %p, state = %u, level = %u\n", dev, state, level));
	if (level == SUSPEND_POWER_DOWN) {
		/* TODO: Need to notify cs_do_magic that it needs to quit ASAP.
		 * Then we must WAIT here until it's finished, before returning
		 * from this function, or else the tasklet will be killed in
		 * the middle of its work. */
		sound_poweroff();
		iis_suspend();
		if(iis_attached_ops.suspend) iis_attached_ops.suspend();
		sound_dma_suspend();
	}
	return 0;
}

static int sound_resume(struct device *dev, u32 level)
{
	D(printk("dev = %p, level = %u\n", dev, level));
	if (level == RESUME_POWER_ON) {
		sound_poweron();
		iis_resume();
		if(iis_attached_ops.resume) iis_attached_ops.resume();
		sound_dma_resume();
	}
	return 0;
}

#else /* CONFIG_PM */
#define sound_suspend NULL
#define sound_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver sound_driver = {
	.name		= "s3c2410-iis",
	.bus		= &platform_bus_type,
	.probe		= sound_probe,
	.remove		= sound_remove,
	.shutdown	= sound_shutdown,
	.suspend	= sound_suspend,
	.resume		= sound_resume,
};

static int __init sound_mod_init(void)
{
	int ret;

	fasync_queue = NULL;
	do_filter = 0;
		
	printk(KERN_INFO "CoolSound engine, (C) 2004 TomTom BV\n");
	D(printk("Registering driver\n"));
	ret = driver_register(&sound_driver);
	if (ret) {
		printk("Unable to register driver (%d)\n", ret);
		return ret;
	}
	D(printk("Done\n"));
	return 0;
}

static void __exit sound_mod_exit(void)
{
	D(printk("Unregistering driver\n"));
	driver_unregister(&sound_driver);
	D(printk("Done\n"));
}

module_init(sound_mod_init);
module_exit(sound_mod_exit);

MODULE_AUTHOR("Thomas Kleffel <tk@maintech.de> ((c) 2005 TomTom BV)");
MODULE_DESCRIPTION("CoolSound engine");
MODULE_LICENSE("GPL");
