#include <linux/types.h>
#include <asm/arch/hardware.h>
#include <asm/arch/regs-gpio.h>

#include "driver.h"
#include "wm8711.h"
#include "spi.h"

static uint16_t wm8711[] = {
	0x0,	/* R0 */
	0x0,	/* R1 */
	0x180,	/* R2 */
	0x0,	/* R3 */
	0x10,	/* R4 */
	0x0,	/* R5 */
	0x60,	/* R6, Power down the clock and the osscilator. */
	0x02,	/* R7 */
#ifdef USE_FS256
	0x20,	/* R8 */
#else /* USE_FS256 */
	0x22,	/* R8 */
#endif /* USE_FS256 */
	0x01	/* R9 */
};

static inline void WM8711_WRREG(uint16_t reg, uint16_t value)
{
	if (reg <= 9)
		wm8711[reg] = value;
	D(printk("R%d=0x%x\n", reg, value));
	spi_write_register(reg, value);
}

static inline uint16_t WM8711_RDREG(uint16_t reg)
{
	return wm8711[reg];
}

static void wm8711_init_registers(void)
{
	int i;

	WM8711_WRREG(15, 0);
	for (i = 0; i < sizeof wm8711 / sizeof *wm8711; i++) {
		if (WM8711_RDREG(i)) WM8711_WRREG(i, WM8711_RDREG(i));
	}
	/* Unmute. */
	WM8711_WRREG(5, 0);
}

static int wm8711_set_volume(unsigned int volume)
{
	if (volume > 121) volume = 121;

	D(printk("Setting volume to: %d\n", volume)); 

	WM8711_WRREG(2, 0x180 | volume); 
	return 0;
}

static int wm8711_set_treble(unsigned int treble)
{
	// Treble is not supported, so do nothing	
	return 0; 
}

static int wm8711_set_bass(unsigned int bass)
{
	// Bass is not supported, so do nothing
	return 0; 
}

static int wm8711_set_register(unsigned int data)
{
	unsigned int reg = data >> 9;
	unsigned int value = (0x1ff & data);

	WM8711_WRREG(reg, value);
	return 0; 
}

static void wm8711_powerdown(void)
{
	spi_write_register(15, 0);
}

void wm8711_init(struct iis_attached_ops *ops)
{
	ops->set_volume = wm8711_set_volume;
	ops->set_volume_internal = wm8711_set_volume;
	ops->set_volume_external = wm8711_set_volume;
	ops->set_register = wm8711_set_register;
	ops->resume = wm8711_init_registers;
	ops->suspend = wm8711_powerdown;
	ops->set_treble = wm8711_set_treble;
	ops->set_bass = wm8711_set_bass;

	wm8711_init_registers();
}


