/* drivers/video/tomtomgo.h
 *
 * TomTom Go LCD Controller Frame Buffer Driver
 *
 * Copyright (c) 2005 TomTom BV
 * Author: Christian Daniel <cd@cdaniel.de>
 * Author:Thomas Kleffel <tk@maintech.de>
 * Author:Jeroen Taverne <jeroen.taverne@tomtom.com>
 * Author:Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 */

#ifndef __DRIVERS_VIDEO_TOMTOMGO_H
#define __DRIVERS_VIDEO_TOMTOMGO_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Shadows for LCD controller registers */

enum tomtomgolcd_screen_type {
	STN_4BDS,
	STN_4BSS,
	STN_8BSS,
	TFT_LCD,
};

enum tomtomgolcd_screen_bpp {
	STN_1BPP,
	STN_2BPP,
	STN_4BPP,
	STN_8BPP,
	STN_12BPP,
	TFT_1BPP,
	TFT_2BPP,
	TFT_4BPP,
	TFT_8BPP,
	TFT_16BPP_5551,
	TFT_16BPP_565,
	TFT_24BPP_LSB,
	TFT_24BPP_MSB,
};

struct tomtomgolcd_screen_info {
	enum tomtomgolcd_screen_type type;
	enum tomtomgolcd_screen_bpp bpp;

	u32	vsize, vfpd, vbpd, vspw;
	u32	hsize, hfpd, hbpd, hspw;
	u32	refresh_max, refresh_min;

	u32 	clkval;

	u8	bswp, hwswp;
	u8 	invvclk, invvline, invvframe, invvd, invvden;
	u8	enpwren, invpwren;
	u8	enlend, invlend;
};

struct tomtomgolcd_fb_slot {
	struct fb_info		fbi;
	struct device		*dev;

	/* raw memory addresses */
	dma_addr_t		map_dma;	/* physical */
	u_char *		map_cpu;	/* virtual */
	u_int			map_size;

	u_int reg_lcdcon1;
	u_int reg_lcdcon2;
	u_int reg_lcdcon3;
	u_int reg_lcdcon4;
	u_int reg_lcdcon5;

	u_int reg_lcdsaddr1;
	u_int reg_lcdsaddr2;
	u_int reg_lcdsaddr3;
	u_int reg_lpc_en;

	u32 pseudo_pal[16];
};

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __DRIVERS_VIDEO_TOMTOMGO_H */

/* EOF */
