#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/mtd/mtd.h>
#include <asm/uaccess.h>
#include <barcelona/debug.h>
#include "ngffs.h"
#include "chunk.h"

/* Defines */
#define PFX "ngffs: "
#define PK_DBG PK_DBG_NONE


extern void ngffs_create_root (struct ngffs_info * ngsb , struct ngffs_block *block);


static u_long erase_done_flags=0;


void hexdump(char *buff) {
	    int i;
	    char bla[2]="\0\0";
	    char bla2[20]="\0";
mdelay(500);
	    for(i=0;i<512;i++) {
		      if(i && !(i%16)) {
			          printk(" - %s\n",bla2);
				      bla2[0]=0;
				        }
		        printk("%02x ",*( ((unsigned char*)buff)+i ));
			  bla[0]=(*( ((unsigned char*)buff)+i))>=32?
				            *( ((unsigned char*)buff)+i):'.';
			    strcat(bla2,bla);
	    }
	        printk(" - %s\n",bla2);
mdelay(500);
}

// TODO what if non-word-aligned read/write?? will mtd handle this for us??
int ngffs_write(struct mtd_info *mtd,__u32 ofs,u_char *buff,int len) {
    int retlen;
    int ret;

    PK_DBG("write at ofs %ul, len=%ul\n",ofs,len);
#if 0
    if( ((struct generic_chunk *)buff)->type==NGFFS_DATA_CHUNK ) {
      hexdump(buff);
    }
#endif

    
    ret=mtd->write(mtd,ofs+CONFIG_NGFFS_START_OFS,len,&retlen,buff);
    if(ret) {
      PK_ERR("ngffs_write: Write error1 at 0x%08x: %d\n", ofs, ret);
      return ret;
    }
    if (retlen != len) {
      PK_ERR("Short write: 0x%x bytes at 0x%08x instead of requested %x\n",
		                   retlen, ofs, len);
      return -EIO;
    }
    return 0;
}

int ngffs_absolute_read(struct mtd_info *mtd,__u32 ofs,u_char *buff,int len) {
    int retlen;
    int ret;

    PK_DBG("read at ofs %ul, len=%ul\n",ofs-CONFIG_NGFFS_START_OFS,len);

    ret=mtd->read(mtd,ofs,len,&retlen,buff);
    if (ret) {
      PK_ERR("ngffs_read: Read error at 0x%08x: %d\n", ofs, ret);
      return ret;
    }
    if (retlen != len) {
      PK_ERR(KERN_NOTICE "Short read: 0x%x bytes at 0x%08x instead of requested %x\n",
             retlen, ofs, len);
      return -EIO;
    }
#if 0
    if( ((struct generic_chunk *)buff)->type==NGFFS_DATA_CHUNK ) {
      hexdump(buff);
    }
#endif
    return 0;
}

int ngffs_read(struct mtd_info *mtd,__u32 ofs,u_char *buff,int len) {
  return ngffs_absolute_read(mtd,ofs+CONFIG_NGFFS_START_OFS,buff,len);
}

void ngffs_erase_callback(struct erase_info *instr) {
  if(instr->state != MTD_ERASE_DONE) {
    PK_DBG("Erase completed, but state is not MTD_ERASE_DONE!\n");
  } else {
		// TODO: spinlock!!
    erase_done_flags|=instr->priv;
		// TODO: spinunlock!!
  } 

  kfree(instr); 
}

int ngffs_erase_flash(struct mtd_info *mtd,int block,u_long mask) {
  struct erase_info *instr;
  int ret;

  instr = kmalloc(sizeof(struct erase_info), GFP_KERNEL);

  memset(instr, 0, sizeof(instr));
  
  instr->mtd=mtd;
  instr->addr=(mtd->erasesize*block)+CONFIG_NGFFS_START_OFS;
  instr->len=mtd->erasesize;
  instr->callback=ngffs_erase_callback;
  instr->priv=mask;

eagain:
  ret = mtd->erase(mtd, instr);
  if (!ret) {
    return 0;
  }
  if(ret==-EROFS) {
    PK_ERR("ngffs: trying to erase read only sector %i\n",block);
    return -EROFS;
  }
  if(ret==-EAGAIN) {
    //schedule();
    goto eagain;
  }
  return ret;
}

int ngffs_erase_ready(u_long mask) {
  return erase_done_flags&mask;
}

// replaces dir chunk in list for the entry_id only if dir->modification is
// larger (TODO: wrap)
// 
// side-effect: invalidates the oldest of the chunks (cleanup), to keep the
// filesystem tidy when power-downs/reboots result in more than one valid 
// chunk for the same entry_id
// 
// returns 1 if replaced or older, 0 if not
int ngffs_replace_older_dir(struct ngffs_info *ngsb,struct dir_chunk *dir,__u32 ofs) {
  struct list_head *curr;

  list_for_each(curr, &ngsb->dirs) {
    struct dir_list_entry *dirle=list_entry(curr, struct dir_list_entry, list);
    if(dirle->chunk->entry_id==dir->entry_id) {
      if(dirle->chunk->modification <= dir->modification) { // added for gc: equals
		// invalidate older chunk
	ngffs_invalidate_chunk(ngsb->sb,dirle->ofs); // (cleanup)
		// replace
	kfree(dirle->chunk);
	dirle->chunk=dir;
	dirle->ofs=ofs;
        return 1;
      } else {
		// older
		// invalidate this chunk
	ngffs_invalidate_chunk(ngsb->sb,ofs);  // (cleanup)
	kfree(dir);
	return 1;
      }
    }
  }
  return 0;
}

// replaces dir chunk in list for the entry_id only if dir->modification is
// larger (TODO: wrap)
// 
// side-effect: invalidates the oldest of the chunks (cleanup), to keep the
// filesystem tidy when power-downs/reboots result in more than one valid 
// chunk for the same entry_id
// 
// returns 1 if replaced or older, 0 if not (ie. insert!)
int ngffs_replace_older_file(struct ngffs_info *ngsb,struct file_chunk *file,__u32 ofs) {
  struct list_head *curr;

  list_for_each(curr, &ngsb->files) {
    struct file_list_entry *filele=list_entry(curr, struct file_list_entry, list);
    if(filele->chunk->entry_id==file->entry_id) {
      if(filele->chunk->modification <= file->modification) { // added for gc: equals
		// invalidate older chunk
	ngffs_invalidate_chunk(ngsb->sb,filele->ofs);	// (cleanup)
		// drop old chunk
	kfree(filele->chunk);
	filele->chunk=file;
	filele->ofs=ofs;
        return 1;
      } else {
		// invalidate this chunk
	ngffs_invalidate_chunk(ngsb->sb,ofs);  // (cleanup)
	kfree(file);
	return 1;
      }
    }
  }
  return 0;
}

int ngffs_scan_block(struct ngffs_info *ngsb, int b) {
  struct mtd_info *mtd=ngsb->mtd;
  struct ngffs_block *blocks=ngsb->blocklist;
  __u32 currofs = EMPTY_BYTES + blocks[b].offset;
  __u32 maxofs= blocks[b].offset+blocks[b].size;
  __u32 freespace=blocks[b].size-EMPTY_BYTES;
  __u32 currsize;
  int done=0;
  int mod_init;
  struct generic_chunk prebuff;
  struct dir_list_entry *dir;
  struct file_list_entry *file;
  struct dir_chunk *tmpdir;
  struct file_chunk *tmpfile;
  int rv;

  while(!done) {
	  // TODO: fix -1 to return valid error codes
    if((rv=ngffs_read(mtd,currofs,(u_char*)&prebuff,sizeof(struct generic_chunk)))) {
	    PK_DBG("ngffs_scan_block(sb,%d): unable to read\n",b);
	    return rv;
    }
    if(prebuff.type==NGFFS_UNUSED) {
      done=1;
    } else {
      currsize=CHUNK_SIZE(&prebuff);
      PK_DBG("inloop> currsize=%ul, ofs=%ul\n",currsize,currofs);
      if(currsize==-1) return -EIO;	// TODO: invalidate block/fs ?
      freespace-=currsize;
      PK_DBG("inloop> |-> freespace=%ul, valid=%x, type=%i\n",freespace,prebuff.valid,prebuff.type);
      if(prebuff.valid==NGFFS_CHUNK_VALID) {
	switch(prebuff.type) {
	  case NGFFS_DIR_CHUNK:
		tmpdir=kmalloc(CHUNK_SIZE(&prebuff),GFP_KERNEL);
		if(tmpdir==NULL) {
		  PK_DBG("ngffs_scan_block(sb,%d): unable to kmalloc dir chunk\n",b);
		  return -ENOMEM;
		}
    		if((rv=ngffs_read(mtd,currofs,(u_char *)(tmpdir),CHUNK_SIZE(&prebuff)))) {
	    	  PK_DBG("ngffs_scan_block(sb,%d): unable to read dir chunk\n",b);
		  return rv;
    		}
		if(!ngffs_replace_older_dir(ngsb,tmpdir,currofs)) {
		  dir=kmalloc(sizeof(struct dir_list_entry),GFP_KERNEL);
		  if(dir==NULL) {
		    PK_DBG("ngffs_scan_block(sb,%d): unable to kmalloc dir_list_entry\n",b);
		    return -ENOMEM;
		  }
		  dir->ofs=currofs;
		  dir->chunk=tmpdir;
		  list_add_tail(&dir->list,&ngsb->dirs);
		}
		if(tmpdir->entry_id > ngsb->max_entry_id) ngsb->max_entry_id=tmpdir->entry_id;
		PK_DBG("ngffs_scan_block(sb,%d): read dir inode %d\n",b,tmpdir->entry_id);
		break;
	  case NGFFS_FILE_CHUNK:
			// alloc mem and read the file chunk
		tmpfile=kmalloc(CHUNK_SIZE(&prebuff),GFP_KERNEL);
		if(tmpfile==NULL) {
		  PK_ERR("ngffs_scan_block(sb,%d): unable to kmalloc file chunk\n",b);
		  return -ENOMEM;
		}
    		if((rv=ngffs_read(mtd,currofs,(u_char *)(tmpfile),CHUNK_SIZE(&prebuff)))) {
	    	  PK_ERR("ngffs_scan_block(sb,%d): unable to read file chunk\n",b);
		  return rv;
    		}
			// is this a later version (this.modified > cached.modified?)
		if(!ngffs_replace_older_file(ngsb,tmpfile,currofs)) {
			// no, alloc new list head
		  file=kmalloc(sizeof(struct file_list_entry),GFP_KERNEL);
		  if(file==NULL) {
		    PK_ERR("ngffs_scan_block(sb,%d): unable to kmalloc file_list_entry\n",b);
		    return -ENOMEM;
		  }
		  file->cache=kmalloc(NGFFS_NUMCHUNKS(NGFFS_MAXFSIZE)*sizeof(struct ngffs_data_cache),
				  								GFP_KERNEL);
		  if(file->cache==NULL) {
		    kfree(file);
		    PK_ERR("ngffs_scan_block(sb,%d): unable to kmalloc file_list_entry.chunkofs\n",b);
		    return -ENOMEM;
		  }
		  	// memset??
		  for(mod_init=0;mod_init<NGFFS_NUMCHUNKS(NGFFS_MAXFSIZE);mod_init++) {
		    file->cache[mod_init].modification=0;
		  }
		  file->ofs=currofs;
		  file->chunk=tmpfile;
		  list_add_tail(&file->list,&ngsb->files);
		}
		if(tmpfile->entry_id > ngsb->max_entry_id) ngsb->max_entry_id=tmpfile->entry_id;
		PK_DBG("ngffs_scan_block(sb,%d): read file inode %d (len=%d)\n",b,tmpfile->entry_id,
										tmpfile->len);
		break;
	  default:
		// ignore all, assume CHUNK_SIZE already caught
		// invalid types
		break;
	}
      }
      currofs+=currsize;
      if(currofs>=maxofs) {
	      PK_WARN("Warning, chunk went outside erase block\n");
	      return -EIO;
      }
    }
  }
  blocks[b].freespace=freespace;
  PK_DBG("blocks[%d].freespace=%d\n",b,freespace);
  return 0;
}


/*
 * Does the second pass of flash scanning, storing offsets for
 * data chunks in the filelisthead's chunk offset cache
 */
int ngffs_build_data_cache(struct ngffs_info * ngsb) {
  struct mtd_info *mtd=ngsb->mtd;
  struct ngffs_block *blocks=ngsb->blocklist;
  int block;
  int ofs;
  struct generic_chunk prebuff;
  struct data_chunk data;

  mtd=ngsb->mtd;
  blocks=ngsb->blocklist;

  for(block=0;block<ngsb->blocks;block++) {
    PK_DBG("second pass scan block %i\n",block);
    if( (blocks[block].header.valid==NGFFS_CHUNK_VALID) && !(blocks[block].header.activated)) {
      ofs=EMPTY_BYTES+sizeof(struct header_chunk);
      while(ofs<blocks[block].size) {
        if(ngffs_read(mtd,blocks[block].offset+ofs,(u_char*)&prebuff,sizeof(struct generic_chunk))) {
          PK_WARN("ngffs_build_data_cache() unable to read from block %i\n",block);
        }
        PK_DBG("at %ul found type %i\n",blocks[block].offset+ofs,prebuff.type);
	switch(prebuff.type) {
	  case NGFFS_UNUSED: ofs=blocks[block].size; break; // abort while loop
	  case NGFFS_DATA_CHUNK:
	    if(prebuff.valid==NGFFS_CHUNK_VALID) {          // only read&cache valid chunks
              if(ngffs_read(mtd,blocks[block].offset+ofs,(u_char*)&data,sizeof(struct data_chunk))) {
                PK_WARN("ngffs_build_data_cache() unable to read from block %i\n",block);
	        return -EIO;
              }
	      ngffs_cache_data_chunk(ngsb,data.entry_id,data.seq_id,data.modification,blocks[block].offset+ofs);
              PK_DBG("data with entry_id=%ul, seq_id=%ul, mod=%ul\n",data.entry_id,data.seq_id,data.modification);
	    }
	    ofs+=CHUNK_SIZE(&prebuff);
	    break;
	  default:
	    ofs+=CHUNK_SIZE(&prebuff);
	    break;
	}
      }
    } else {
	PK_DBG("second pass, block %i invalid or not activated\n",block);
    }
  }
  return 0;
}



int ngffs_scan_flash(struct ngffs_info *ngsb) {
  int activeblocks=0;
  int b;
  size_t retlen;
  int ret,i;
  int erase=0;
  u_char buff[EMPTY_BYTES+sizeof(struct header_chunk)];
  u_char empty[EMPTY_BYTES];
  struct header_chunk *header=(struct header_chunk *)(buff+EMPTY_BYTES);
  struct mtd_info *mtd;
  struct ngffs_block *blocks;
  int numblocks;
  int rv;

  mtd=ngsb->mtd;
  blocks=ngsb->blocklist;
  numblocks=ngsb->blocks;

	// TODO memset?
  for(i=0;i<EMPTY_BYTES;i++) empty[i]=0;

  PK_DBG("ngffs: scanning %d blocks\n",numblocks);
restart:
  activeblocks=0;
  erase=0;
  for(b=0;b<numblocks;b++) {
recheck:
    PK_DBG("ngffs: scanning block %d\n",b);
    ret=mtd->read(mtd,(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES+sizeof(struct header_chunk),&retlen,buff);
    if (ret) {
      PK_ERR("ngffs_scan_flash: Read error at 0x%08x: %d\n", mtd->erasesize*b, ret);
      return ret;
    }
    if (retlen != (sizeof(struct header_chunk)+EMPTY_BYTES)) {
      PK_ERR("Short read: 0x%x bytes at 0x%08x instead of requested %x\n",
             retlen, mtd->erasesize*b, sizeof(struct header_chunk)+EMPTY_BYTES);
      return -EIO;
    }
    PK_DBG("header->valid = 0x%x, header->type = %i\n",header->valid,header->type);

    erase=(header->valid==NGFFS_CHUNK_VALID)?0:1;
  
    for(i=0;i<EMPTY_BYTES && !erase;i++) {
      if(buff[i]!=0) erase++;
    }

    if(erase) {
	PK_DBG("ngffs: erasing sector %i\n",b);

	ngffs_erase_flash(mtd,b,1);
        while(!ngffs_erase_ready(1)) {
	  // schedule();
	}
	PK_DBG("write at 0x%x, num=%i\n",(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES*sizeof(u_char));
	ret=mtd->write(mtd,(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES*sizeof(u_char),&retlen,empty);
        if(ret) {
          PK_ERR("ngffs_scan_flash: Write error2 at 0x%08x: %d\n", mtd->erasesize*b, ret);
          return ret;
        }
        if (retlen != (EMPTY_BYTES*sizeof(u_char))) {
          PK_ERR(KERN_NOTICE "Short write: 0x%x bytes at 0x%08x instead of requested %x\n",
                 retlen, mtd->erasesize*b, EMPTY_BYTES*sizeof(u_char));
          return -EIO;
        }

	goto recheck;	// todo: some retry count
    }
    if( !header->activated ) {  // implicit: header->valid==0xff
      activeblocks++;
    }
    blocks[b].offset = mtd->erasesize * b;
    blocks[b].size = mtd->erasesize;
    blocks[b].freespace = -1;	// will be filled by ngffs_scan_block

    bcopy(header,&(blocks[b].header),sizeof(struct header_chunk));

    rv=ngffs_scan_block(ngsb,b);
    if(rv==-ENOMEM) return -ENOMEM;       // out of memory errors don't mean that the fs is corrupt
    if(rv) { 				  // other errors do mean corrupted fs, so reformat 
      for(b=0;b<numblocks;b++) {
	PK_DBG("ngffs: erasing sector %i\n",b);

	ngffs_erase_flash(mtd,b,1);
        while(!ngffs_erase_ready(1)) {
	  // schedule();
	}
	PK_DBG("write at 0x%x, num=%i\n",(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES*sizeof(u_char));
	ret=mtd->write(mtd,(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES*sizeof(u_char),&retlen,empty);
        if(ret) {
          PK_ERR("ngffs_scan_flash: Write error2 at 0x%08x: %d\n", mtd->erasesize*b, ret);
          return ret;
        }
        if (retlen != (EMPTY_BYTES*sizeof(u_char))) {
          PK_ERR(KERN_NOTICE "Short write: 0x%x bytes at 0x%08x instead of requested %x\n",
                 retlen, mtd->erasesize*b, EMPTY_BYTES*sizeof(u_char));
          return -EIO;
        }
      }
      goto restart;	// and restart scanning // TODO: retry count
    }
  }
	// is this actually the first time (ie. did we do a format just now?)
  if(activeblocks==0) {
	// make all blocks active, except the first one

    for(i=1;i<numblocks;i++) {
      PK_DBG("ngffs: activating block %i\n",i);

      	// read header
      ret=mtd->read(mtd,(blocks[i].offset)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES+sizeof(struct header_chunk),&retlen,buff);
      if (ret) {
        PK_ERR("ngffs_scan_flash: Read error at 0x%08x: %d\n", blocks[i].offset, ret);
        return ret;
      }
      if (retlen != (sizeof(struct header_chunk)+EMPTY_BYTES)) {
        PK_ERR("Short read: 0x%x bytes at 0x%08x instead of requested %x\n",
             retlen, blocks[i].offset, sizeof(struct header_chunk)+EMPTY_BYTES);
        return -EIO;
      }

      header->type=NGFFS_HEADER_CHUNK;
      header->activated=0;
      header->maxfsize= NGFFS_MAXFSIZE;
      header->maxcsize= NGFFS_MAXCSIZE;
      header->maxdirent= NGFFS_MAXDIRENT;

      ret=mtd->write(mtd,(blocks[i].offset)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES+sizeof(struct header_chunk),&retlen,buff);
      if(ret) {
        PK_ERR("ngffs_scan_flash: Write error3 at 0x%08x: %d\n", blocks[i].offset, ret);
        return ret;
      }
      if (retlen != (EMPTY_BYTES+sizeof(struct header_chunk))) {
        PK_ERR("Short write: 0x%x bytes at 0x%08x instead of requested %x\n",
             retlen, blocks[i].offset, EMPTY_BYTES+sizeof(struct header_chunk));
        return -EIO;
      }
      blocks[i].freespace = mtd->erasesize - ( EMPTY_BYTES+sizeof(struct header_chunk) );

//    blocks[0].offset = 0; bullshit, already done above..
//    blocks[0].size = mtd->erasesize; dito

      //PK_DBG("copy from 0x%x to 0x%x (len=%d)\n",header,&(blocks->header),sizeof(struct header_chunk));

      bcopy(header,&(blocks[i].header),sizeof(struct header_chunk));
    }
    ngffs_create_root(ngsb,&(blocks[1]));
    ngsb->max_entry_id=99;	// start at 100, <100 is for sysfile subsystem
  } else {	// activeblocks>0, there may be data in there
    if((ret=ngffs_build_data_cache(ngsb))) {
      PK_DBG("ngffs_build_data_cache failed..\n");
      return ret;
    }
  }
  PK_DBG("ngffs: scan done\n");
  return 0;
}

/*
**  Flash scan routine for already mounted filesystems (used after
**  garbage collection to update the various caches)
*/
int ngffs_rescan_flash(struct super_block * sb) {
  int b;
  size_t retlen;
  int ret;
  u_char buff[EMPTY_BYTES+sizeof(struct header_chunk)];
  struct header_chunk *header=(struct header_chunk *)(buff+EMPTY_BYTES);
  struct ngffs_info *ngsb;
  struct mtd_info *mtd;
  struct ngffs_block *blocks;
  int numblocks;

  ngsb=NGFFS_INFO(sb);
  mtd=ngsb->mtd;
  blocks=ngsb->blocklist;
  numblocks=ngsb->blocks;

  for(b=0;b<numblocks;b++) {
    PK_DBG("ngffs: re-scanning block %d\n",b);
    ret=mtd->read(mtd,(mtd->erasesize*b)+CONFIG_NGFFS_START_OFS,EMPTY_BYTES+sizeof(struct header_chunk),&retlen,buff);
    if (ret) {
      PK_ERR(KERN_NOTICE "ngffs_scan_flash: Read error at 0x%08x: %d\n", mtd->erasesize*b, ret);
      return ret;
    }
    if (retlen != (sizeof(struct header_chunk)+EMPTY_BYTES)) {
      PK_ERR("Short read: 0x%x bytes at 0x%08x instead of requested %x\n",
             retlen, mtd->erasesize*b, sizeof(struct header_chunk)+EMPTY_BYTES);
      return -EIO;
    }
 	// overwrite current header for this block
    bcopy(header,&(blocks[b].header),sizeof(struct header_chunk));

	// is this a valid & active block? 
    if(header->valid && !(header->activated)) {
	PK_DBG("block is valid & active\n");
		// yes => scan it
		// note that this will overwrite file & dir chunks but
		// leaves dir list entries and file list entries intact
      ngffs_scan_block(ngsb,b);
    } else {
	PK_DBG("block is NOT valid & active\n");
    }
  }

  if((ret=ngffs_build_data_cache(ngsb))) {
    PK_ERR("ngffs_build_data_cache failed..\n");
    return ret;
  }

  return 0;
}
