#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/blkdev.h>
#include <linux/mtd/mtd.h>
#include <linux/time.h>
#include <asm/uaccess.h>
#include <asm/errno.h>
#include <barcelona/debug.h>
#include "ngffs.h"
#include "chunk.h"
#include "file.h"
#include "flash.h"
#include "garbage.h"

#define NGFFS_SYSFILE

#include "sysfile.h"
#include "sysfiles.h"

/* Defines */
#define PFX "ngffs: "
#define PK_DBG PK_DBG_FUNC


int ngffs_sysfile_readpage (struct file *filp, struct page *pg);


struct file_operations ngffs_sysfile_operations =
{
        llseek:         generic_file_llseek,
        open:           generic_file_open,
        read:           generic_file_read,
        write:          generic_file_write,
//        ioctl:          ngffs_ioctl,
        mmap:           generic_file_mmap,
//        fsync:          ngffs_null_fsync
};

/* ngffs_file_inode_operations */

struct inode_operations ngffs_sysfile_inode_operations =
{
//        setattr:        ngffs_setattr
};


struct address_space_operations ngffs_sysfile_address_operations =
{
        readpage:       ngffs_sysfile_readpage,
//        prepare_write:  ngffs_prepare_write,
//        commit_write:   ngffs_commit_write
};



int ngffs_sysfile_do_readpage_nolock(struct inode *inode, struct page *pg) {
  struct ngffs_info *ngsb=NGFFS_INFO(inode->i_sb);
  int i;
  int rv=0;
  
  i=inode->i_ino-3;
  PK_DBG("sysfile found at %i\n",i);

  PK_DBG("sysfile read");
  if (!PageLocked(pg))
    PAGE_BUG(pg);

  if(i>=NGFFS_SYSFILES) {
    PK_WARN("sysfile id out of range!\n");
    goto readpage_fail;
  }

/*
struct ngffs_sysfile {
  __u32 length;
  __u32 ofs;
  char *name;
};
*/
  rv=ngffs_absolute_read(ngsb->mtd,ngffs_sysfiles[i].ofs,(u_char *)page_address(pg),ngffs_sysfiles[i].length);
  if(rv) goto readpage_fail;
 
  SetPageUptodate(pg);
  ClearPageError(pg);
  flush_dcache_page(pg);
  kunmap(pg);
  return 0;

readpage_fail:
  ClearPageUptodate(pg);
  SetPageError(pg);
  kunmap(pg);
  return rv;

}

int ngffs_sysfile_readpage (struct file *filp, struct page *pg) {

  PK_DBG("ngffs_sysfile_readpage()\n");
	// do lock stuff?
	ngffs_sysfile_do_readpage_nolock(filp->f_dentry->d_inode,pg);
        unlock_page(pg);
	// unlock
  return 0;
}


