/* include/barcelona/Barc_Gpio.h
 *
 * Public interface for the GPIO driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef __INCLUDE_BARCELONA_BARC_GPIO_H
#define __INCLUDE_BARCELONA_BARC_GPIO_H

#ifndef __INCLUDE_BARCELONA_TYPES_H
#include <barcelona/types.h>
#endif /* __INCLUDE_BARCELONA_TYPES_H */

#include <linux/ioctl.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define GPIO_DEVNAME					"hwstatus"
#define GPIO_MAJOR						240

/* For IOR_HWSTATUS ioctl */
typedef struct {
	UINT8 u8ChargeStatus;
	UINT8 u8InputStatus;
} HARDWARE_STATUS;

/* For HARDWARE_STATUS::u8ChargeStatus */
#define CHARGE_STATE_ACPWR_OFF			0		/* AC power is off   */
#define CHARGE_STATE_FAULT				1		/* Charge fault      */
#define CHARGE_STATE_COMPLETE			2		/* Charging complete */
#define CHARGE_STATE_FAST_CHARGING		3		/* Fast charging     */
#define CHARGE_STATE_TOPPINGUP			4		/* Topping up        */

/* For HARDWARE_STATUS::u8InputStatus */
#define ONOFF_MASK       0x01	/* On-off button      */
#define USB_DETECT_MASK  0x02	/* USB power detect   */
#define DOCK_SENSE0_MASK 0x04
#define DOCK_SENSE1_MASK 0x08
#define DOCK_LIGHTS_MASK 0x10

/* For IOW_OUT_PIN_STATUS ioctl */
#define MUTE_EXT_SPEAKER_MASK			0x01	/* External speaker  */

/* GPIO driver ioctls */
#define GPIO_DRIVER_MAGIC				'U'
#define IOR_HWSTATUS					_IOR(GPIO_DRIVER_MAGIC, 0, HARDWARE_STATUS)
#define IOW_MUTE_ON_EXTERNAL_SPEAKER	_IO(GPIO_DRIVER_MAGIC, 1)
#define IOW_MUTE_OFF_EXTERNAL_SPEAKER	_IO(GPIO_DRIVER_MAGIC, 2)
#define IOW_USB_OFF						_IO(GPIO_DRIVER_MAGIC, 5)
#define IOW_USB_ON						_IO(GPIO_DRIVER_MAGIC, 6)
#define IOR_DISK_ACCESS       _IOR(GPIO_DRIVER_MAGIC, 7, UINT32)

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __INCLUDE_BARCELONA_BARC_GPIO_H */

/* EOF */
